@@ -1,7 +1,8 @@
 # HOST_TEST for espefuse.py using the pytest framework
 #
 # Supports esp32, esp32s2, esp32s3beta2, esp32s3,
-#          esp32c3, esp32h2beta1, esp32c2, esp32c6
+#          esp32c3, esp32h2beta1, esp32c2, esp32c6, esp32p4,
+#          esp32c61, esp32c5, esp32c5beta3.
 #
 # How to use:
 #
@@ -72,6 +73,7 @@ class EfuseTestCase:
                 f"{sys.executable} -m espefuse --chip {arg_chip} "
                 f"--virt --path-efuse-file {self.efuse_file.name} -d"
             )
+            self._set_target_wafer_version()
         else:
             self.base_cmd = (
                 f"{sys.executable} -m espefuse --chip {arg_chip} "
@@ -113,6 +115,14 @@ class EfuseTestCase:
     def _set_34_coding_scheme(self):
         self.espefuse_py("burn_efuse CODING_SCHEME 1")
 
+    def _set_none_recovery_coding_scheme(self):
+        self.espefuse_py("burn_efuse CODING_SCHEME 3")
+
+    def _set_target_wafer_version(self):
+        # ESP32 has to be ECO3 (v3.0) for tests
+        if arg_chip == "esp32":
+            self.espefuse_py("burn_efuse CHIP_VER_REV1 1 CHIP_VER_REV2 1")
+
     def check_data_block_in_log(
         self, log, file_path, repeat=1, reverse_order=False, offset=0
     ):
@@ -173,6 +183,18 @@ class TestReadCommands(EfuseTestCase):
         self.espefuse_py("dump -h")
         self.espefuse_py("dump")
 
+    def test_dump_format_joint(self):
+        tmp_file = tempfile.NamedTemporaryFile(delete=False)
+        self.espefuse_py(f"dump --format joint --file_name {tmp_file.name}")
+
+    def test_dump_split_default(self):
+        tmp_file = tempfile.NamedTemporaryFile(delete=False)
+        self.espefuse_py(f"dump --file_name {tmp_file.name}")
+
+    def test_dump_split(self):
+        tmp_file = tempfile.NamedTemporaryFile(delete=False)
+        self.espefuse_py(f"dump --format split --file_name {tmp_file.name}")
+
     def test_summary(self):
         self.espefuse_py("summary -h")
         self.espefuse_py("summary")
@@ -180,6 +202,18 @@ class TestReadCommands(EfuseTestCase):
     def test_summary_json(self):
         self.espefuse_py("summary --format json")
 
+    def test_summary_filter(self):
+        self.espefuse_py("summary MAC")
+        self.espefuse_py("summary --format value_only MAC")
+        self.espefuse_py(
+            "summary --format value_only MAC WR_DIS",
+            check_msg="The 'value_only' format can be used exactly for one efuse.",
+            ret_code=2,
+        )
+
+    @pytest.mark.skipif(
+        arg_chip == "esp32p4", reason="No Custom MAC Address defined yet"
+    )
     def test_get_custom_mac(self):
         self.espefuse_py("get_custom_mac -h")
         if arg_chip == "esp32":
@@ -192,6 +226,17 @@ class TestReadCommands(EfuseTestCase):
         self.espefuse_py("adc_info -h")
         self.espefuse_py("adc_info")
 
+    def test_adc_info_2(self):
+        if arg_chip == "esp32":
+            self.espefuse_py("burn_efuse BLK3_PART_RESERVE 1")
+        elif arg_chip in ["esp32c3", "esp32s3", "esp32s3beta2"]:
+            self.espefuse_py("burn_efuse BLK_VERSION_MAJOR 1")
+        elif arg_chip in ["esp32c2", "esp32s2", "esp32c6"]:
+            self.espefuse_py("burn_efuse BLK_VERSION_MINOR 1")
+        elif arg_chip in ["esp32h2", "esp32h2beta1"]:
+            self.espefuse_py("burn_efuse BLK_VERSION_MINOR 2")
+        self.espefuse_py("adc_info")
+
     def test_check_error(self):
         self.espefuse_py("check_error -h")
         self.espefuse_py("check_error")
@@ -277,9 +322,14 @@ class TestReadProtectionCommands(EfuseTestCase):
                 ret_code=2,
             )
         else:
+            key1_purpose = (
+                "USER"
+                if arg_chip in ["esp32p4", "esp32c61", "esp32c5", "esp32c5beta3"]
+                else "RESERVED"
+            )
             self.espefuse_py(
                 f"burn_key BLOCK_KEY0 {IMAGES_DIR}/256bit USER \
-                BLOCK_KEY1 {IMAGES_DIR}/256bit RESERVED \
+                BLOCK_KEY1 {IMAGES_DIR}/256bit {key1_purpose} \
                 BLOCK_KEY2 {IMAGES_DIR}/256bit SECURE_BOOT_DIGEST0 \
                 BLOCK_KEY3 {IMAGES_DIR}/256bit SECURE_BOOT_DIGEST1 \
                 BLOCK_KEY4 {IMAGES_DIR}/256bit SECURE_BOOT_DIGEST2 \
@@ -343,9 +393,13 @@ class TestWriteProtectionCommands(EfuseTestCase):
             efuse_lists = """RD_DIS DIS_DOWNLOAD_ICACHE
                            XTS_KEY_LENGTH_256 UART_PRINT_CONTROL"""
             efuse_lists2 = "RD_DIS DIS_DOWNLOAD_ICACHE"
+        elif arg_chip == "esp32p4":
+            efuse_lists = """RD_DIS KEY_PURPOSE_0 SECURE_BOOT_KEY_REVOKE0
+                           SPI_BOOT_CRYPT_CNT"""
+            efuse_lists2 = "RD_DIS KEY_PURPOSE_0 KEY_PURPOSE_2"
         else:
             efuse_lists = """RD_DIS DIS_ICACHE DIS_FORCE_DOWNLOAD
-                           DIS_CAN SOFT_DIS_JTAG DIS_DOWNLOAD_MANUAL_ENCRYPT
+                           DIS_DOWNLOAD_MANUAL_ENCRYPT
                            USB_EXCHG_PINS WDT_DELAY_SEL SPI_BOOT_CRYPT_CNT
                            SECURE_BOOT_KEY_REVOKE0 SECURE_BOOT_KEY_REVOKE1
                            SECURE_BOOT_KEY_REVOKE2 KEY_PURPOSE_0 KEY_PURPOSE_1
@@ -353,11 +407,16 @@ class TestWriteProtectionCommands(EfuseTestCase):
                            SECURE_BOOT_EN SECURE_BOOT_AGGRESSIVE_REVOKE FLASH_TPUW
                            DIS_DOWNLOAD_MODE
                            ENABLE_SECURITY_DOWNLOAD UART_PRINT_CONTROL
-                           MAC OPTIONAL_UNIQUE_ID
+                           MAC
                            BLOCK_USR_DATA BLOCK_KEY0 BLOCK_KEY1
                            BLOCK_KEY2 BLOCK_KEY3 BLOCK_KEY4 BLOCK_KEY5"""
-            if arg_chip not in ["esp32h2", "esp32h2beta1"] and arg_chip not in [
-                "esp32c6"
+            if arg_chip not in [
+                "esp32h2",
+                "esp32h2beta1",
+                "esp32c6",
+                "esp32c61",
+                "esp32c5",
+                "esp32c5beta3",
             ]:
                 efuse_lists += """ DIS_DOWNLOAD_ICACHE
                             SPI_PAD_CONFIG_CLK SPI_PAD_CONFIG_Q
@@ -380,6 +439,7 @@ class TestWriteProtectionCommands(EfuseTestCase):
             )
 
 
+@pytest.mark.skipif(arg_chip == "esp32p4", reason="No Custom MAC Address defined yet")
 class TestBurnCustomMacCommands(EfuseTestCase):
     def test_burn_custom_mac(self):
         self.espefuse_py("burn_custom_mac -h")
@@ -431,8 +491,13 @@ class TestBurnCustomMacCommands(EfuseTestCase):
 
 
 @pytest.mark.skipif(
-    arg_chip in ["esp32c2", "esp32h2beta1", "esp32c3", "esp32c6", "esp32h2"],
-    reason=f"TODO: add support set_flash_voltage for {arg_chip}",
+    arg_chip
+    not in [
+        "esp32",
+        "esp32s2",
+        "esp32s3",
+    ],
+    reason=f"{arg_chip} does not support set_flash_voltage",
 )
 class TestSetFlashVoltageCommands(EfuseTestCase):
     def test_set_flash_voltage_1_8v(self):
@@ -499,7 +564,7 @@ class TestSetFlashVoltageCommands(EfuseTestCase):
         )
 
 
-@pytest.mark.skipif(arg_chip != "esp32c3", reason="Not necessary fo all chips")
+@pytest.mark.skipif(arg_chip != "esp32c3", reason="Not necessary for all chips")
 class TestValueArgForBurnEfuseCommands(EfuseTestCase):
     def test_efuse_is_bool_given_none(self):
         self.espefuse_py("burn_efuse SECURE_BOOT_KEY_REVOKE0")
@@ -617,6 +682,9 @@ class TestBurnEfuseCommands(EfuseTestCase):
         assert "(Override SD_CMD pad (GPIO11/SPICS0)) 0b00000 -> 0b11111" in output
         assert "BURN BLOCK0  - OK (all write block bits are set)" in output
 
+    @pytest.mark.skipif(
+        arg_chip == "esp32p4", reason="No Custom MAC Address defined yet"
+    )
     def test_burn_mac_custom_efuse(self):
         crc_msg = "(OK)"
         self.espefuse_py("burn_efuse -h")
@@ -665,23 +733,21 @@ class TestBurnEfuseCommands(EfuseTestCase):
                 SECURE_BOOT_EN 1 \
                 UART_PRINT_CONTROL 1"
             )
-            self.espefuse_py(
-                "burn_efuse \
-                OPTIONAL_UNIQUE_ID 0x2328ad5ac9145f698f843a26d6eae168",
-                check_msg="-> 0x2328ad5ac9145f698f843a26d6eae168",
-            )
-            output = self.espefuse_py("summary -d")
-            assert (
-                "read_regs: d6eae168 8f843a26 c9145f69 2328ad5a "
-                "00000000 00000000 00000000 00000000"
-            ) in output
-            assert "= 68 e1 ea d6 26 3a 84 8f 69 5f 14 c9 5a ad 28 23 R/W" in output
-            efuse_from_blk2 = "BLK_VERSION_MAJOR"
-            if arg_chip == "esp32s2":
-                efuse_from_blk2 = "BLK_VERSION_MINOR"
-            if arg_chip != "esp32c6":
+            if arg_chip not in ["esp32c5", "esp32c5beta3", "esp32c61"]:
+                # chips having the OPTIONAL_UNIQUE_ID field
                 self.espefuse_py(
-                    f"burn_efuse {efuse_from_blk2} 1",
+                    "burn_efuse \
+                    OPTIONAL_UNIQUE_ID 0x2328ad5ac9145f698f843a26d6eae168",
+                    check_msg="-> 0x2328ad5ac9145f698f843a26d6eae168",
+                )
+                output = self.espefuse_py("summary -d")
+                assert (
+                    "read_regs: d6eae168 8f843a26 c9145f69 2328ad5a "
+                    "00000000 00000000 00000000 00000000"
+                ) in output
+                assert "= 68 e1 ea d6 26 3a 84 8f 69 5f 14 c9 5a ad 28 23 R/W" in output
+                self.espefuse_py(
+                    "burn_bit BLOCK_SYS_DATA 1",
                     check_msg="Burn into BLOCK_SYS_DATA is forbidden "
                     "(RS coding scheme does not allow this).",
                     ret_code=2,
@@ -753,6 +819,24 @@ class TestBurnEfuseCommands(EfuseTestCase):
             ADC2_TP_HIGH 45"
         )
 
+    @pytest.mark.skipif(
+        arg_chip != "esp32s3",
+        reason="Currently S3 only has this efuse incompatibility check",
+    )
+    def test_burn_efuse_incompatibility_check(self):
+        self.espefuse_py(
+            "burn_efuse DIS_USB_JTAG 1 DIS_USB_SERIAL_JTAG 1",
+            check_msg="Incompatible eFuse settings detected, abort",
+            ret_code=2,
+        )
+        self.espefuse_py("burn_efuse DIS_USB_JTAG 1")
+        self.espefuse_py(
+            "burn_efuse DIS_USB_SERIAL_JTAG 1",
+            check_msg="Incompatible eFuse settings detected, abort",
+            ret_code=2,
+        )
+        self.espefuse_py("burn_efuse DIS_USB_SERIAL_JTAG 1 --force")
+
 
 class TestBurnKeyCommands(EfuseTestCase):
     @pytest.mark.skipif(arg_chip != "esp32", reason="ESP32-only")
@@ -862,6 +946,10 @@ class TestBurnKeyCommands(EfuseTestCase):
             "esp32h2beta1",
             "esp32c6",
             "esp32h2",
+            "esp32p4",
+            "esp32c5",
+            "esp32c5beta3",
+            "esp32c61",
         ],
         reason="Only chips with 6 keys",
     )
@@ -870,9 +958,13 @@ class TestBurnKeyCommands(EfuseTestCase):
                BLOCK_KEY0 {IMAGES_DIR}/256bit   XTS_AES_256_KEY_1 \
                BLOCK_KEY1 {IMAGES_DIR}/256bit_1 XTS_AES_256_KEY_2 \
                BLOCK_KEY2 {IMAGES_DIR}/256bit_2 XTS_AES_128_KEY"
-        if arg_chip in ["esp32c3", "esp32c6"] or arg_chip in [
+        if arg_chip in [
+            "esp32c3",
+            "esp32c6",
             "esp32h2",
             "esp32h2beta1",
+            "esp32c5",
+            "esp32c5beta3",
         ]:
             cmd = cmd.replace("XTS_AES_256_KEY_1", "XTS_AES_128_KEY")
             cmd = cmd.replace("XTS_AES_256_KEY_2", "XTS_AES_128_KEY")
@@ -946,8 +1038,8 @@ class TestBurnKeyCommands(EfuseTestCase):
         self.check_data_block_in_log(output, f"{IMAGES_DIR}/192bit_2")
 
     @pytest.mark.skipif(
-        arg_chip not in ["esp32s2", "esp32s3"],
-        reason="512 bit keys are only supported on ESP32-S2 and S3",
+        arg_chip not in ["esp32s2", "esp32s3", "esp32p4", "esp32c61"],
+        reason="512 bit keys are only supported on ESP32-S2, S3, P4, C61",
     )
     def test_burn_key_512bit(self):
         self.espefuse_py(
@@ -964,11 +1056,11 @@ class TestBurnKeyCommands(EfuseTestCase):
         )
 
     @pytest.mark.skipif(
-        arg_chip not in ["esp32s2", "esp32s3"],
-        reason="512 bit keys are only supported on ESP32-S2 and S3",
+        arg_chip not in ["esp32s2", "esp32s3", "esp32p4", "esp32c61"],
+        reason="512 bit keys are only supported on ESP32-S2, S3, P4, C61",
     )
     def test_burn_key_512bit_non_consecutive_blocks(self):
-        # Burn efuses seperately to test different kinds
+        # Burn efuses separately to test different kinds
         # of "key used" detection criteria
         self.espefuse_py(
             f"burn_key \
@@ -1007,8 +1099,8 @@ class TestBurnKeyCommands(EfuseTestCase):
         ) in output
 
     @pytest.mark.skipif(
-        arg_chip not in ["esp32s2", "esp32s3"],
-        reason="512 bit keys are only supported on ESP32-S2 and S3",
+        arg_chip not in ["esp32s2", "esp32s3", "esp32p4", "esp32c61"],
+        reason="512 bit keys are only supported on ESP32-S2, S3, P4, C61",
     )
     def test_burn_key_512bit_non_consecutive_blocks_loop_around(self):
         self.espefuse_py(
@@ -1039,7 +1131,10 @@ class TestBurnKeyCommands(EfuseTestCase):
             "acadaeaf a8a9aaab a4a5a6a7 22a1a2a3"
         ) in output
 
-    @pytest.mark.skipif(arg_chip != "esp32h2", reason="Only for ESP32-H2 chips")
+    @pytest.mark.skipif(
+        arg_chip not in ["esp32h2", "esp32c5", "esp32c5beta3", "esp32c61", "esp32p4"],
+        reason="These chips support ECDSA_KEY",
+    )
     def test_burn_key_ecdsa_key(self):
         self.espefuse_py(
             f"burn_key \
@@ -1062,7 +1157,10 @@ class TestBurnKeyCommands(EfuseTestCase):
             "00000000 00000000 00000000 00000000"
         ) in output
 
-    @pytest.mark.skipif(arg_chip != "esp32h2", reason="Only for ESP32-H2 chips")
+    @pytest.mark.skipif(
+        arg_chip not in ["esp32h2", "esp32c5", "esp32c5beta3", "esp32c61", "esp32p4"],
+        reason="These chips support ECDSA_KEY",
+    )
     def test_burn_key_ecdsa_key_check_byte_order(self):
         self.espefuse_py(
             f"burn_key \
@@ -1164,6 +1262,10 @@ class TestBurnBlockDataCommands(EfuseTestCase):
             "esp32h2beta1",
             "esp32c6",
             "esp32h2",
+            "esp32p4",
+            "esp32c5",
+            "esp32c5beta3",
+            "esp32c61",
         ],
         reason="Only chip with 6 keys",
     )
@@ -1301,6 +1403,10 @@ class TestBurnBlockDataCommands(EfuseTestCase):
             "esp32h2beta1",
             "esp32c6",
             "esp32h2",
+            "esp32p4",
+            "esp32c5",
+            "esp32c5beta3",
+            "esp32c61",
         ],
         reason="Only chips with 6 keys",
     )
@@ -1496,6 +1602,10 @@ class TestBurnKeyDigestCommandsEsp32C2(EfuseTestCase):
         "esp32h2beta1",
         "esp32c6",
         "esp32h2",
+        "esp32p4",
+        "esp32c5",
+        "esp32c5beta3",
+        "esp32c61",
     ],
     reason="Supports 6 key blocks",
 )
@@ -1607,6 +1717,10 @@ class TestBurnBitCommands(EfuseTestCase):
             "esp32h2beta1",
             "esp32c6",
             "esp32h2",
+            "esp32p4",
+            "esp32c5",
+            "esp32c5beta3",
+            "esp32c61",
         ],
         reason="Only chip with 6 keys",
     )
@@ -1657,6 +1771,16 @@ class TestBurnBitCommands(EfuseTestCase):
             ret_code=2,
         )
 
+    @pytest.mark.skipif(arg_chip != "esp32", reason="ESP32-only")
+    def test_burn_bit_with_none_recovery_coding_scheme(self):
+        self._set_none_recovery_coding_scheme()
+        self.espefuse_py("burn_bit BLOCK3 0 1 2 4 8 16 32 64 96 128 160 192 224 255")
+        self.espefuse_py(
+            "summary",
+            check_msg="17 01 01 00 01 00 00 00 01 00 00 00 01 00 00 "
+            "00 01 00 00 00 01 00 00 00 01 00 00 00 01 00 00 80",
+        )
+
 
 @pytest.mark.skipif(
     arg_chip != "esp32", reason="Tests are only for esp32. (TODO: add for all chips)"
@@ -1729,7 +1853,7 @@ class TestByteOrderBurnKeyCommand(EfuseTestCase):
 class TestExecuteScriptsCommands(EfuseTestCase):
     @classmethod
     def setup_class(self):
-        # Save the current working directory to be resotred later
+        # Save the current working directory to be restored later
         self.stored_dir = os.getcwd()
 
     @classmethod
@@ -1737,14 +1861,20 @@ class TestExecuteScriptsCommands(EfuseTestCase):
         # Restore the stored working directory
         os.chdir(self.stored_dir)
 
-    @pytest.mark.skipif(arg_chip == "esp32c2", reason="TODO: Add tests for esp32c2")
+    @pytest.mark.skipif(
+        arg_chip in ["esp32c2", "esp32p4"],
+        reason="These chips do not have eFuses used in this test",
+    )
     def test_execute_scripts_with_check_that_only_one_burn(self):
         self.espefuse_py("execute_scripts -h")
         name = arg_chip if arg_chip in ["esp32", "esp32c2"] else "esp32xx"
         os.chdir(os.path.join(TEST_DIR, "efuse_scripts", name))
         self.espefuse_py("execute_scripts execute_efuse_script2.py")
 
-    @pytest.mark.skipif(arg_chip == "esp32c2", reason="TODO: Add tests for esp32c2")
+    @pytest.mark.skipif(
+        arg_chip in ["esp32c2", "esp32p4"],
+        reason="These chips do not have eFuses used in this test",
+    )
     def test_execute_scripts_with_check(self):
         self.espefuse_py("execute_scripts -h")
         name = arg_chip if arg_chip in ["esp32", "esp32c2"] else "esp32xx"
@@ -1945,3 +2075,58 @@ class TestKeyPurposes(EfuseTestCase):
             "key due to a hardware bug (please see TRM for more details)",
             ret_code=2,
         )
+
+
+class TestPostponedEfuses(EfuseTestCase):
+    def test_postpone_efuses(self):
+        if arg_chip == "esp32":
+            cmd = f"--postpone \
+                    burn_efuse UART_DOWNLOAD_DIS 1 \
+                    burn_key BLOCK1 {IMAGES_DIR}/256bit \
+                    burn_efuse ABS_DONE_1 1 FLASH_CRYPT_CNT 1"
+            num = 1
+        else:
+            sb_digest_name = (
+                "SECURE_BOOT_DIGEST" if arg_chip == "esp32c2" else "SECURE_BOOT_DIGEST0"
+            )
+            cmd = f"--postpone \
+                burn_efuse ENABLE_SECURITY_DOWNLOAD 1 DIS_DOWNLOAD_MODE 1 \
+                SECURE_VERSION 1 \
+                burn_key BLOCK_KEY0 {IMAGES_DIR}/256bit {sb_digest_name} \
+                burn_efuse SPI_BOOT_CRYPT_CNT 1 SECURE_BOOT_EN 1"
+            num = 3 if arg_chip == "esp32c2" else 4
+        output = self.espefuse_py(cmd)
+        assert f"BURN BLOCK{num}  - OK" in output
+        assert "BURN BLOCK0  - OK" in output
+        assert "Burn postponed efuses from BLOCK0" in output
+        assert "BURN BLOCK0  - OK" in output
+        assert "Successful" in output
+
+
+class TestCSVEfuseTable(EfuseTestCase):
+    def test_extend_efuse_table_with_csv_file(self):
+        csv_file = f"{IMAGES_DIR}/esp_efuse_custom_table.csv"
+        output = self.espefuse_py(f" --extend-efuse-table {csv_file} summary")
+        assert "MODULE_VERSION (BLOCK3)" in output
+        assert "DEVICE_ROLE (BLOCK3)" in output
+        assert "SETTING_2 (BLOCK3)" in output
+        assert "ID_NUM_0 (BLOCK3)" in output
+        assert "ID_NUM_1 (BLOCK3)" in output
+        assert "ID_NUM_2 (BLOCK3)" in output
+        assert "CUSTOM_SECURE_VERSION (BLOCK3)" in output
+        assert "ID_NUMK_0 (BLOCK3)" in output
+        assert "ID_NUMK_1 (BLOCK3)" in output
+
+        self.espefuse_py(
+            f"--extend-efuse-table {csv_file} burn_efuse \
+                         MODULE_VERSION 1 \
+                         CUSTOM_SECURE_VERSION 4 \
+                         SETTING_1_ALT_NAME 7 \
+                         SETTING_2 1 \
+                         ID_NUM_0 1 \
+                         ID_NUM_1 1 \
+                         ID_NUM_2 1 \
+                         MY_ID_NUMK_0 1 \
+                         MY_ID_NUMK_1 1 \
+                         MY_DATA_FIELD1 1"
+        )
