@@ -1,6 +1,7 @@
 import hashlib
 import os
 import os.path
+import re
 import struct
 import subprocess
 import sys
@@ -40,7 +41,7 @@ def segment_matches_section(segment, section):
 class BaseTestCase:
     @classmethod
     def setup_class(self):
-        # Save the current working directory to be resotred later
+        # Save the current working directory to be restored later
         self.stored_dir = os.getcwd()
         os.chdir(TEST_DIR)
 
@@ -113,7 +114,7 @@ class BaseTestCase:
                 f" segment(s) in bin image (image segments: {image.segments})"
             )
 
-    def assertImageInfo(self, binpath, chip="esp8266"):
+    def assertImageInfo(self, binpath, chip="esp8266", assert_sha=False):
         """
         Run esptool.py image_info on a binary file,
         assert no red flags about contents.
@@ -126,7 +127,13 @@ class BaseTestCase:
         except subprocess.CalledProcessError as e:
             print(e.output)
             raise
-        assert "invalid" not in output, "Checksum calculation should be valid"
+        assert re.search(
+            r"Checksum: [a-fA-F0-9]{2} \(valid\)", output
+        ), "Checksum calculation should be valid"
+        if assert_sha:
+            assert re.search(
+                r"Validation Hash: [a-fA-F0-9]{64} \(valid\)", output
+            ), "SHA256 should be valid"
         assert (
             "warning" not in output.lower()
         ), "Should be no warnings in image_info output"
@@ -267,7 +274,11 @@ class TestESP32Image(BaseTestCase):
         try:
             self.run_elf2image("esp32", elfpath, extra_args=extra_args)
             image = esptool.bin_image.LoadFirmwareImage("esp32", binpath)
-            self.assertImageInfo(binpath, "esp32")
+            self.assertImageInfo(
+                binpath,
+                "esp32",
+                True if "--ram-only-header" not in extra_args else False,
+            )
             return image
         finally:
             try_delete(binpath)
@@ -315,13 +326,20 @@ class TestESP32Image(BaseTestCase):
         # this ELF will produce 8 segments in the bin
         image = self._test_elf2image(ELF, BIN)
         # Adjacent sections are now merged, len(image.segments) should
-        # equal 4 (instead of 8).
-        assert len(image.segments) == 4
+        # equal 5 (instead of 8).
+        assert len(image.segments) == 5
 
         # --use_segments uses ELF segments(phdrs), produces just 2 segments in the bin
         image = self._test_elf2image(ELF, BIN, ["--use_segments"])
         assert len(image.segments) == 2
 
+    def test_ram_only_header(self):
+        ELF = "esp32-app-template.elf"
+        BIN = "esp32-app-template.bin"
+        # --ram-only-header produces just 2 visible segments in the bin
+        image = self._test_elf2image(ELF, BIN, ["--ram-only-header"])
+        assert len(image.segments) == 2
+
 
 class TestESP8266FlashHeader(BaseTestCase):
     def test_2mb(self):
@@ -397,7 +415,7 @@ class TestELFSHA256(BaseTestCase):
                     0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff, 0xffffffff},
     };
 
-    This leaves zeroes only for the fiels of SHA-256 and the test will fail
+    This leaves zeroes only for the fields of SHA-256 and the test will fail
     if the placement of zeroes are tested at the wrong place.
 
     00000000: e907 0020 780f 0840 ee00 0000 0000 0000  ... x..@........
